from dataclasses import MISSING, Field, fields
from datetime import date
from re import fullmatch
from typing import ClassVar

from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from .custom_types import int_pk, samplesheet_str_pk, stripped_str_pk

__all__ = ['Base', 'Entity', 'Data', 'Process']


class Base(MappedAsDataclass, DeclarativeBase):
    @classmethod
    def fields(cls) -> tuple[Field, ...]:
        return fields(cls)

    @classmethod
    def init_fields(cls) -> tuple[Field, ...]:
        return tuple(f for f in cls.fields() if f.init)

    @classmethod
    def required_init_fields(cls) -> tuple[Field, ...]:
        return tuple(
            f
            for f in fields(cls)
            if f.init and f.default is MISSING and f.default_factory is MISSING
        )

    @classmethod
    def field_names(cls) -> set[str]:
        return {f.name for f in cls.fields()}

    @classmethod
    def init_field_names(cls) -> set[str]:
        return {f.name for f in cls.init_fields()}

    @classmethod
    def required_init_field_names(cls) -> set[str]:
        return {f.name for f in cls.required_init_fields()}

    pass


class Entity(Base, kw_only=True):
    __abstract__ = True

    id: Mapped[int_pk] = mapped_column(repr=False, compare=False, default=None)


class Data(Base, kw_only=True):
    __abstract__ = True

    # TODO: auto incrementing behavior
    id: Mapped[samplesheet_str_pk]

    # Model metadata
    id_date_col: ClassVar[str]
    id_prefix: ClassVar[str]
    id_length: ClassVar[int]

    def __post_init__(self):
        self.id = self.id.strip().upper()

        date_col: date = getattr(self, self.id_date_col)
        year_last_two_digits = date_col.strftime('%y')

        prefix = self.id_prefix + year_last_two_digits
        suffix_length = self.id_length - len(prefix)
        pattern = rf'{prefix}\d{{{suffix_length}}}'

        model_name = type(self).__name__

        if fullmatch(pattern, self.id) is None:
            pass
            # raise ValueError(
            #     f'{model_name} ID {self.id} does not match the pattern {pattern}.'
            # )


class Process(Base, kw_only=True):
    __abstract__ = True

    name: Mapped[stripped_str_pk]



from datetime import date
from typing import ClassVar, Literal

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..bases import Data
from ..custom_types import samplesheet_str, stripped_str
from .entities import Lab, Person
from .processes import Assay, Platform

__all__ = ['DataSet', 'Sample']


class DataSet(Data, kw_only=True):
    __tablename__ = 'data_set'

    # DataSet attributes
    name: Mapped[samplesheet_str] = mapped_column(index=True)
    ilab_request_id: Mapped[stripped_str] = mapped_column(
        index=True
    )  # TODO: ilab request ID validation
    date_initialized: Mapped[date] = mapped_column(repr=False)

    # Parent foreign keys
    assay_name: Mapped[str] = mapped_column(
        ForeignKey('assay.name'), repr=False, default=None
    )
    lab_id: Mapped[int] = mapped_column(
        ForeignKey('lab.id'), default=None, repr=False, init=False
    )
    platform_name: Mapped[str] = mapped_column(ForeignKey('platform.name'), init=False)
    submitter_id: Mapped[int] = mapped_column(
        ForeignKey('person.id'), default=None, repr=False, init=False
    )

    # Parent models
    lab: Mapped[Lab] = relationship()
    submitter: Mapped[Person] = relationship()

    # Model metadata
    id_date_col: ClassVar[Literal['date_initialized']] = 'date_initialized'

    __mapper_args__ = {
        'polymorphic_on': 'platform_name',
    }

    def __post_init__(self):
        self.name = self.lab.name + self.assay_name + self.date_initialized.isoformat()


class Sample(Data, kw_only=True):
    __tablename__ = 'sample'

    # Sample attributes
    name: Mapped[samplesheet_str] = mapped_column(index=True)
    date_received: Mapped[date] = mapped_column(repr=False)

    # Parent foreign keys
    data_set_id: Mapped[int] = mapped_column(
        ForeignKey('data_set.id'), default=None, repr=False, init=False
    )
    platform_name: Mapped[str] = mapped_column(ForeignKey('platform.name'), init=False)

    # Model metadata
    id_date_col: ClassVar[Literal['date_received']] = 'date_received'

    __mapper_args__ = {'polymorphic_on': 'platform_name'}
